#include "config.h"

module gpu_solver_class
  use iso_c_binding
  use comlib
  use lattice_class
  use counter_class
  use field_gauge_class
  use field_fermion_class
  use field_clover_class
  use gpu_bind_class
  use single_field_class
  implicit none
  private

  public :: gpu_solver
  public :: gpu_solver_bicgstab
  public :: gpu_solver_cg
  public :: new, delete
  public :: get_name
  public :: assign_inv_mult_eoprec_wd
  public :: assign_inv_mult_Qee_shift

  real(DP), parameter :: MAX_TOL = 0.01_DP
  real(DP), parameter :: MIN_TOL = 0.00001_DP

  logical, save :: m_is_gpu_initialized   = .false.

  type gpu_solver
    private
    type(s_wqf_eo_obj),  allocatable :: swe
    type(s_wqf_eo_obj),  allocatable :: sve
    type(s_gvf_obj),     allocatable :: su
#ifdef _CLOVER
    type(s_clvf_obj),    allocatable :: sfcl
#endif
    real(DP), public :: err_norm
    real(DP), public :: tol
    integer,  public :: iter
    integer,  public :: print_flag = 0
    integer,  public :: ieo
  end type

  type, extends(gpu_solver) :: gpu_solver_bicgstab
    character(len=CHARLEN), private :: name = "GPU-BICGStab"
  end type

  type, extends(gpu_solver) :: gpu_solver_cg
    character(len=CHARLEN), private :: name = "GPU-CG"
  end type


  interface new
    module procedure new_gpu_solver_f
  end interface

  interface delete
    module procedure delete_gpu_solver_f
  end interface

  interface get_name
    module procedure get_name_gpu_solver
  end interface

  interface assign_inv_mult_eoprec_wd
    module procedure assign_inv_mult_eoprec_wd_gpu
  end interface

  interface assign_inv_mult_Qee_shift
    module procedure assign_inv_mult_Qee_shift_blk_gpu
  end interface


  interface 
   !
   ! Fortran interfaces for GPU code using F2003 C binding
   !

#ifdef _CLOVER
    subroutine new_gpu_solver_clover(u, fclinv) BIND(C,name="new_gpu_solver_clover_")
      import s_gvf_obj
      import s_clvf_obj
      implicit none
      type(s_gvf_obj),  intent(in) :: u
      type(s_clvf_obj), intent(in) :: fclinv
    end subroutine
#endif

    subroutine new_gpu_solver_wilson(u) BIND(C,name="new_gpu_solver_wilson_")
      import s_gvf_obj
      implicit none
      type(s_gvf_obj),  intent(in) :: u
    end subroutine

    subroutine delete_gpu_solver(print_flag) BIND(C,name="delete_gpu_solver_")
      import C_INT
      implicit none
      integer(C_INT), intent(in) :: print_flag
    end subroutine

    subroutine s_bicgstab_gpu(kappa, tol, iter, dye, ye) BIND(C,name="cuda_assign_inv_mult_eoprec_wd_bicgstab_")
      import c_double, c_int, s_wqf_eo_obj
      implicit none
      real(C_DOUBLE),     intent(in)    :: tol, kappa
      integer(C_INT),     intent(in)    :: iter
      type(s_wqf_eo_obj), intent(inout) :: ye, dye
    end subroutine

    subroutine s_cg_gpu(kappa, tol, iter, dye, ye) BIND(C,name="cuda_assign_inv_mult_eoprec_wd_cg_")
      import c_double, c_int, s_wqf_eo_obj
      implicit none
      real(C_DOUBLE),     intent(in)    :: tol, kappa
      integer(C_INT),     intent(in)    :: iter
      type(s_wqf_eo_obj), intent(inout) :: ye, dye
    end subroutine

    subroutine assign_inv_mult_Qee_shift_cg(kappa, tol, iter, dye, ye, shift) BIND(C,name="cuda_assign_inv_mult_Qee_shift_cg_")
      import c_double, c_int, s_wqf_eo_obj
      implicit none
      real(C_DOUBLE),     intent(in)    :: tol, kappa, shift
      integer(C_INT),     intent(in)    :: iter
      type(s_wqf_eo_obj), intent(inout) :: ye, dye
    end subroutine

    subroutine assign_inv_mult_Qee_cg(kappa, tol, iter, dye, ye) BIND(C,name="cuda_assign_inv_mult_Qee_cg_")
      import c_double, c_int, s_wqf_eo_obj
      implicit none
      real(C_DOUBLE),     intent(in)    :: tol, kappa
      integer(C_INT),     intent(in)    :: iter
      type(s_wqf_eo_obj), intent(inout) :: ye, dye
    end subroutine

  end interface
  
contains

#ifdef _CLOVER
subroutine new_gpu_solver_f(this,u,ieo,fclinv)
#else
subroutine new_gpu_solver_f(this,u,ieo)
#endif
  use error_class
  implicit none
  class(gpu_solver),      intent(inout) :: this
  type(vfield_gluon_wg),  intent(in) :: u
  integer,                intent(in) :: ieo
#ifdef _CLOVER
  type(field_clover_term), optional, intent(in) :: fclinv
#endif
  logical :: is_clover
  integer :: num_gpus, num_use_gpus
  type(single_field_world) :: sp_lattice

#ifdef _DEBUG
  this%print_flag = 1
  if ( 0 == nodeid ) write(*,'("%%Start new_gpu_solver_f")')
#else
  this%print_flag = 0
#endif

  if (ieo < 0 .or. ieo > 1) then
    call error_stop("new_gpu_solver_f: ieo error.")
  endif
  this%ieo = ieo

#ifdef _CLOVER
  if (present(fclinv)) then
    is_clover = .true.
  else
    is_clover = .false.
  endif
#else
  is_clover = .false.
#endif

  call new(sp_lattice)
  num_use_gpus=NGX
  call bind_cpu_to_gpu(num_gpus,num_use_gpus)
  gpu_mult_count(:) = 0

  this%iter = 0

  if (allocated(this%swe)) deallocate(this%swe)
  if (allocated(this%sve)) deallocate(this%sve)
  if (allocated(this%su))  deallocate(this%su)
  allocate(this%swe)
  allocate(this%sve)
  allocate(this%su)
  call new(this%su)

#ifdef _CLOVER
  if (is_clover) then
    if (allocated(this%sfcl)) deallocate(this%sfcl)
    allocate(this%sfcl)
    call new(this%sfcl)
  endif
#endif

  call conv_u(this%su,u)
#ifdef _CLOVER
  if (is_clover) then
    call conv_clv(this%sfcl,fclinv)
  endif
#endif

  if (is_clover) then
#ifdef _CLOVER
    call new_gpu_solver_clover(this%su,this%sfcl)
#endif
  else
    call new_gpu_solver_wilson(this%su)
  endif

  if (allocated(this%su)) deallocate(this%su)
#ifdef _CLOVER
  if (is_clover) then
    if (allocated(this%sfcl)) deallocate(this%sfcl)
  endif
#endif
  m_is_gpu_initialized = .true.

#ifdef _DEBUG_
  if ( 0 == nodeid ) write(*,'("%%End new_gpu_solver_f")')
#endif

  return
end subroutine

subroutine delete_gpu_solver_f(this)
  implicit none
  class(gpu_solver),      intent(inout) :: this

  this%iter = 0
  gpu_mult_count(:) = 0

  if (allocated(this%swe))  deallocate(this%swe)
  if (allocated(this%sve))  deallocate(this%sve)
  if (allocated(this%su))   deallocate(this%su)
#ifdef _CLOVER
  if (allocated(this%sfcl)) deallocate(this%sfcl)
#endif

  if (m_is_gpu_initialized) then
    call delete_gpu_solver(this%print_flag)
  endif

  m_is_gpu_initialized = .false.

  return
end subroutine

function get_name_gpu_solver(this) result(name)
  class(gpu_solver), intent(in) :: this
  character(len=CHARLEN) :: name
  select type(this)
  class is (gpu_solver_bicgstab)
    name = this%name
  class is (gpu_solver_cg)
    name = this%name
  end select
  return
end function

subroutine assign_inv_mult_eoprec_wd_gpu(this,kappa,dye,ye)
!
!       v
! dye = Dee \ ye
!
  use error_class
  implicit none
  class(gpu_solver), intent(inout) :: this
  real(DP),          intent(in)    :: kappa
  complex(DP),       intent(inout) ::  ye(:)  ! packed data
  complex(DP),       intent(inout) :: dye(:)  ! packed data
  character(len=CHARLEN) :: str
  integer  :: ieo
  real(DP) :: norm, tol

  ieo = this%ieo
  gpu_mult_count(:) = 0
  tol = MAX(MIN(this%tol/this%err_norm/10.0_DP,MAX_TOL),MIN_TOL)

  call new(this%swe,ieo)
  !
  ! swe <= ye/|ye|, norm = |ye|
  !
  call unpack(ye,this%swe,norm)
!write(*,*)abs2(this%swe)

  !        v
  ! sve <= Dee \ swe
  !
  select type(this)
  class is (gpu_solver_bicgstab)

    call s_bicgstab_gpu(kappa, tol, this%iter, this%sve, this%swe)

  class is (gpu_solver_cg)

    call       s_cg_gpu(kappa, tol, this%iter, this%sve, this%swe)

  end select
!  this%sve = this%swe

  !
  ! dye <= sve*norm
  !
  call pack(this%sve,dye,norm)    

  call inc(mult_iter,gpu_mult_count(1))
  gpu_mult_count(:) = 0

  !==========================
  ! check solver convergence
  !==========================
  if (this%iter < 0) then
    call delete(this)
    write(str,'("assign_inv_mult_eoprec_wd_gpu : ",A," at ",I5)')__FILE__,__LINE__
    call error_message(TRIM(str))
    call error_message("assign_inv_mult_eoprec_wd_gpu : maximam itertion reached.")
    call error_message("CONTINUE.")
  endif

  return
end subroutine

subroutine assign_inv_mult_Qee_shift_blk_gpu(this,kappa,dye,ye,shift)
!
!          
! dye(j) = (Qee - shift(j)) \ ye(j)
!
!        v       v
!  Qee = Dee^dag Dee
!
!
  use error_class
  implicit none
  class(gpu_solver), intent(inout) :: this
  real(DP),          intent(in)    :: kappa
  type(field_quark_eo_wg), intent(inout) :: dye(:)
  type(field_quark_eo_wg), intent(inout) ::  ye(:)
  real(DP), optional, intent(in)         :: shift(:)
  character(len=CHARLEN) :: str
  logical :: is_shift
  integer :: ieo, num_pf, ipf
  real(DP) :: norm, tol

  is_shift = present(shift)
  num_pf = SIZE(ye(:))

  ieo = this%ieo
  gpu_mult_count(:) = 0

  call new(this%swe,ieo)

  do ipf = 1,num_pf
    !
    ! swe <= ye/|ye|, norm = |ye|
    !
    call conv_q_d2s(this%swe,ye(ipf),norm)
 
    tol = MAX(MIN(this%tol/this%err_norm/10.0_DP,MAX_TOL),MIN_TOL)

    !        
    ! sve <= (Qee - shift(j)) \ swe
    !
    if (is_shift) then
      call assign_inv_mult_Qee_shift_cg(kappa, tol, this%iter, this%sve, this%swe, shift(ipf))
    else
      call assign_inv_mult_Qee_cg(      kappa, tol, this%iter, this%sve, this%swe)
    endif

    call inc(mult_iter,gpu_mult_count(ipf))

    !
    ! dye <= sve*norm
    !
    call conv_q_s2d(dye(ipf),this%sve,norm)    
!    call pack(this%sve,dye(ipf),norm)    

  enddo

  gpu_mult_count(:) = 0

  !==========================
  ! check solver convergence
  !==========================
  if (this%iter < 0) then
    call delete(this)
    write(str,'("assign_inv_mult_Qee_shift_blk_gpu : ",A," at ",I5)')__FILE__,__LINE__
    call error_message(TRIM(str))
    call error_message("assign_inv_mult_Qee_shift_blk_gpu : maximam itertion reached.")
    call error_message("CONTINUE.")
  endif

  return
end subroutine

end module
