#include "../config.h"

!
!= Definition of Quark Propagator
!
module quark_prop_class
!
!= This defines quark propagtor for measurement program
!
!  The quark parameter is defined as a pointer.
!  The wavefunction for source and sink are defined as pointers. 
!  These pointers should be assigned or initialized before calling get_qprop.
!  User should set the source (numerical) data on quark propagator before calling get_qprop.
!
!
!  set_source is defined for typical wavefunction setting, 
!  with whith the numerical data are generated on a quark propagator according 
!  to the wavefunction source data.
!
!
  use comlib
  use lattice_class
  use timer_class
  use logfile_class
  use field_gauge_class
  use field_fermion_class
#ifdef _CLOVER
  use quark_clover_class
#else
  use quark_wilson_class
#endif
  use quark_dwf_class
  use quark_dw_ovlp_class
  use wavefunc_class
  implicit none
  private
  public :: quark_prop
  public :: qprop_params
  public :: new
  public :: delete
  public :: read
  public :: print
  public :: print_statistics
  public :: get_prop
  public :: clear_prop
  public :: store
  public :: restore
  public :: assign
  public :: assign_mult
  public :: set_source
  public :: assign_inv_mult_wd
!  public :: conv_prop_to_y
!  public :: conv_y_to_prop

  integer, parameter :: NSPACE=NX*NY*NZ

  type qprop_params
  !
  ! quark propagator parameter for physics measurement
  !
    class(quark_wilson),  pointer   :: mass => NULL()
    type(timer)                     :: etime
    type(logfile), pointer          :: solver_log => NULL()
    integer                         :: quark_type = 0
    character(len=CHARLEN)          :: fname
    character(len=CHARLEN)          :: path
    real(DP)                        :: tol
!    integer                         :: smear_sous ! 0 for local, 1 for smeared
!    integer                         :: smear_sink ! 0 for local, 1 for smeared
  end type


  type quark_prop
  !
  ! quark propagator for physics measurement
  !
    complex(DP),        allocatable :: prop(:,:,:,:,:,:) ! prop(sink,sous,sink,sous,sink,sink)
    real(DP)                        :: psum(NTT)         ! check sum
    real(DP)                        :: plq               ! as config check sum
    type(qprop_params),     pointer :: parameter     => NULL()
    type(src_wavefunc_obj), pointer :: wavefunc_sous => NULL()
    type(snk_wavefunc_obj), pointer :: wavefunc_sink => NULL()
    integer                         :: smear_sous ! 0 for local, 1 for smeared
    integer                         :: smear_sink ! 0 for local, 1 for smeared
    integer                         :: itraj = 0
    integer                         :: idummy = 0
  end type

  interface new
    module procedure new_qprop
    module procedure new_qprop_params
  end interface

  interface delete
    module procedure delete_qprop
    module procedure delete_qprop_params
  end interface

  interface read
    module procedure read_qprop
    module procedure read_qprop_params
  end interface

  interface print
    module procedure print_qprop
    module procedure print_qprop_params
  end interface

  interface assign
    module procedure assign_qprop
    module procedure assign_qprop_params
  end interface

  interface get_prop
    module procedure get_qprop
  end interface

  interface clear_prop
    module procedure clear_qprop
  end interface

  interface store
    module procedure store_qprop
    module procedure store_qprop_params
  end interface

  interface restore
    module procedure restore_qprop
    module procedure restore_qprop_params
  end interface

  interface print_statistics
    module procedure print_stat_qprop
  end interface

  interface assign_mult
    module procedure assign_mult_u_q
    module procedure assign_mult_u_y
  end interface

contains

subroutine new_qprop(this)
!
! set default values for a quark propagator
!
  implicit none
  type(quark_prop), intent(inout) :: this
  if (allocated(this%prop)) deallocate(this%prop)
  this%parameter     => NULL()
  this%wavefunc_sous => NULL()
  this%wavefunc_sink => NULL()
  this%itraj  = 0
  this%idummy = 0
  return
end subroutine

subroutine delete_qprop(this)
!
! set default values for a quark propagator
!
  implicit none
  type(quark_prop), intent(inout) :: this
  if (allocated(this%prop)) deallocate(this%prop)
  this%parameter     => NULL()
  this%wavefunc_sous => NULL()
  this%wavefunc_sink => NULL()
  this%itraj  = 0
  this%idummy = 0
  return
end subroutine

subroutine new_qprop_params(this)
!
! set default values for a quark propagator parameter
!
  implicit none
  type(qprop_params), intent(inout) :: this

  this%solver_log  => NULL()
  this%tol = 10*EPSILON(1.0_DP)
  call new(this%etime)

  return
end subroutine

subroutine delete_qprop_params(this)
!
! delete quark parameter parameters
!
  implicit none
  type(qprop_params), intent(inout) :: this
  this%solver_log => NULL()
  select type(p => this%mass)
#ifdef _CLOVER
  type is (quark_clover)
#else
  type is (quark_wilson)
#endif
    call delete(p)
  type is (quark_truncd_dw_overlap)
    call delete(p)
  end select
  return
end subroutine

subroutine assign_qprop_params(param1,param2)
!
! assignment parameters
!
! param1 := param2
!
  type(qprop_params), intent(inout) :: param1
  type(qprop_params), intent(in)    :: param2
  select type(p1 => param1%mass)
#ifdef _CLOVER
  type is (quark_clover)
#else
  type is (quark_wilson)
#endif

    select type(p2 => param2%mass)
#ifdef _CLOVER
    type is (quark_clover)
#else
    type is (quark_wilson)
#endif
      call assign(p1,p2)
    end select

  type is (quark_truncd_dw_overlap)

    select type(p2 => param2%mass)
    type is (quark_truncd_dw_overlap)
      call assign(p1,p2)
    end select

  end select

  param1%solver_log  => param2%solver_log
  param1%fname       =  param2%fname
  param1%path        =  param2%path
!  param1%smear_sous  =  param2%smear_sous
!  param1%smear_sink  =  param2%smear_sink
  return
end subroutine

subroutine read_qprop(this,iout)
!
! read parameters for a quark propagator
!
  implicit none
  type(quark_prop), intent(inout) :: this
  integer,     intent(in)    :: iout
  call read(this%parameter,iout)
  return
end subroutine

subroutine read_qprop_params(this,iout)
!
! read parameters for a quark propagator parameter
!
  use error_class
  implicit none
  type(qprop_params), intent(inout) :: this
  integer,            intent(in)    :: iout
  character(len=CHARLEN) :: str
  real(DP) :: kappa,csw
  integer, parameter :: Nfone=1

  !
  ! read wilson quark mass parameters
  !

  if (nodeid==0) then
    !
    ! quark type
    !
    read(iout,*)this%quark_type
  endif

  if (NPU > 1) then
    call comlib_bcast(this%quark_type,0)
  endif

  select case(this%quark_type)
  case (0)

#ifdef _CLOVER
    allocate(quark_clover::this%mass)
#else
    allocate(quark_wilson::this%mass)
#endif

  case (1)

    allocate(quark_truncd_dw_overlap::this%mass)

  case default

    write(str,'("Quark type must be 0(Wilson/Clover) or 1(DWOVLP).")')
    call error_stop(str)

  end select

  select type(p => this%mass)
#ifdef _CLOVER
  type is (quark_clover)
#else
  type is (quark_wilson)
#endif

    call new(p,0)
    call read(p,iout)

  type is (quark_truncd_dw_overlap)

    call new(p,0)
    call read(p,iout)
    call set_coef(p)

  end select

  if (nodeid==0) then
    !
    ! read Solver parameters
    !
    read(iout,*)this%tol
  endif

  if (NPU > 1) then
    call comlib_bcast(this%tol,0)
  endif

  return
end subroutine

subroutine print_qprop(this)
!
! print parameters for a quark propagator
!
  implicit none
  type(quark_prop), intent(inout) :: this

  call print(this%parameter)
  if (nodeid == 0) then
    write(*,'("==== Smearing type ====")')
    write(*,'(3X,"          Source :",I3)')this%smear_sous
    write(*,'(3X,"            Sink :",I3)')this%smear_sink
  endif
  if (nodeid==0) then
    write(*,'("==== Source Wavefunction type ====")')
  endif
  call print(this%wavefunc_sous)

!  if (0 < this%parameter%smear_sink) then
  if (0 < this%smear_sink) then
    if (nodeid==0) then
      write(*,'("==== Sink Wavefunction type ====")')
    endif
    call print(this%wavefunc_sink)
  endif

  if (nodeid==0) then
    write(*,'(80("="))')
  endif

  return
end subroutine

subroutine print_qprop_params(this)
!
! print parameters for a quark propagator
!
  implicit none
  type(qprop_params), intent(inout) :: this

  if (nodeid==0) then
    write(*,'("==== Mass Params ====")')
    select type(p => this%mass)
#ifdef _CLOVER
    type is (quark_clover)
#else
    type is (quark_wilson)
#endif
      call print(p)
    type is (quark_truncd_dw_overlap)
      call print(p)
    end select
  endif

  if (nodeid==0) then
    write(*,'("==== Solver Params ====")')
    write(*,'(3X," Solver tol      :",E24.16)')this%tol
  endif

  return
end subroutine

subroutine psum(this)
!
! compute check sum of a quark propagator (sum=pion propagator).
!
  implicit none
  type(quark_prop), intent(inout) :: this
  real(DP) :: rtmp
  integer :: it,itt,ic,is,jc,js,isp

  this%psum(:) = 0.0_DP

  do it=1,NT
!    itt = it + ipsite(4)*NT
    itt = it
    rtmp= 0.0_DP
!$OMP PARALLEL DO PRIVATE(isp,ic,is,jc,js) REDUCTION(+:rtmp)
    do isp=1,NSPACE
      do js=1,SPIN
      do is=1,SPIN
      do jc=1,COL
      do ic=1,COL
        rtmp=rtmp+ real(this%prop(ic,jc,is,js,isp,it))**2 &
 &               +aimag(this%prop(ic,jc,is,js,isp,it))**2
      enddo
      enddo
      enddo
      enddo
    enddo
    this%psum(itt)=rtmp
  enddo

#ifndef _singlePU
  call comlib_sumcast(this%psum)
#endif

  return
end subroutine

subroutine print_stat_qprop(this)
!
! print out quark propagator statistics
!
  implicit none
  type(quark_prop), intent(inout) :: this
  integer :: itt
  real(DP) :: etime

  call psum(this)

  etime = get_elapse(this%parameter%etime)
  if (nodeid==0) then
    write(*,'("==== Solver Statistics ====")')
    write(*,'(3X,"             Time :",F12.4)')etime
  endif

  if (nodeid==0) then
    write(*,'("==== Quark Prop |S(itt)|^2 ====")')
    do itt=1,NTT
      write(*,'(I3,E24.16)')itt,this%psum(itt)
    enddo
    write(*,'("")')
  endif

  return
end subroutine

subroutine clear_qprop(this)
!
! clear quark propagator data
!
  implicit none
  type(quark_prop), intent(inout) :: this
!$OMP PARALLEL WORKSHARE
  this%prop(:,:,:,:,:,:) = Z0
!$OMP END PARALLEL WORKSHARE
  return
end subroutine

subroutine assign_qprop(qp1,qp2)
!
! assign(=duplicate) quark propagator (incuding parameters and data)
!
! q1 := q2
!
  implicit none
  type(quark_prop), intent(inout) :: qp1
  type(quark_prop), intent(in)    :: qp2
  if (allocated(qp1%prop)) deallocate(qp1%prop)
  allocate(qp1%prop(COL,COL,SPIN,SPIN,NX*NY*NZ,NTT))
  qp1%prop = qp2%prop
  qp1%psum = qp2%psum
  qp1%parameter     => qp2%parameter
  qp1%wavefunc_sous => qp2%wavefunc_sous
  qp1%wavefunc_sink => qp2%wavefunc_sink
  return
end subroutine

subroutine set_source(this)
!
! set source on a quark propagator via wavefunction
!
  use error_class
  use print_status_class
  implicit none
  type(quark_prop), intent(inout) :: this
  integer :: it,ic,is,isp,jc,js
  integer :: ix,iy,iz,itx,ity,itz,itt
  character(len=CHARLEN) :: str
  complex(DP) :: ctmp

  if (allocated(this%prop)) deallocate(this%prop)
  allocate(this%prop(COL,COL,SPIN,SPIN,NX*NY*NZ,NTT))

  if (.not.associated(this%wavefunc_sous)) then
    call error_stop("set_source: Error, assign wavefunc_sous for qprop.")
  endif

!$OMP PARALLEL DO PRIVATE(it,isp,ic,is,jc,js)
  do it=1,NT
  do isp=1,NSPACE
    do js=1,SPIN
    do is=1,SPIN
    do jc=1,COL
    do ic=1,COL
      this%prop(ic,jc,is,js,isp,it) = Z0
    enddo
    enddo
    enddo
    enddo
    do is=1,SPIN
    do ic=1,COL
      this%prop(ic,ic,is,is,isp,it) = this%wavefunc_sous%wavefunc(ic,is,isp,it)
    enddo
    enddo
  enddo
  enddo

!#define _DDD_
#ifdef _DDD_
!==============
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
  do it=1,NT
    itx = ix + ipsite(1)*NX
    ity = iy + ipsite(2)*NY
    itz = iz + ipsite(3)*NZ
    itt = it
    isp = ispace(ix,iy,iz)
    do js=1,SPIN
    do jc=1,COL
    do is=1,SPIN
    do ic=1,COL
      ctmp = this%prop(ic,jc,is,js,isp,it)
      call comlib_sumcast(ctmp)
      if (ABS(ctmp) > 0.0_DP) then
        write(str,'(4I4,2I2,2I2,2E24.15," SETSRC")')& 
 &       itx,ity,itz,itt,ic,is,jc,js, &
 &       this%prop(ic,jc,is,js,isp,it)
         call print_status(str)
      endif
    enddo
    enddo
    enddo
    enddo
  enddo
  enddo
  enddo
  enddo
!==============
#endif

  return
end subroutine

subroutine conv_prop_to_y(this,y,ic0,is0)
!
! conversion from quark propagator to fermion field
!
!    this : quark propagator
!       y : fermion field
! ic0,is0 : source color-spinor of the quark propagator to be converted to fermion field.
!
  implicit none
  type(quark_prop),     intent(in)    :: this
  type(field_quark_wg), intent(inout) :: y
  integer, intent(in) :: ic0,is0
  integer :: ieoxyz,ith,ic,is,ieo,isp
  integer :: ix,iy,iz,it

  do ieo=0,1
    y%eo(ieo)%ieo = ieo
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,ith,it,isp,is,ic)
    do ix=1,NX  ! site loop in 1 block.
    do iy=1,NY
    do iz=1,NZ
    ieoxyz=mod(ipeo+iz+iy+ix+ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      it=ith*2+ieoxyz
      isp=ispace(ix,iy,iz)

      do is=1,SPIN
      do ic=1,COL
        y%eo(ieo)%s(ith,iz,iy,ix)%y(ic,is) = this%prop(ic,ic0,is,is0,isp,it)
      enddo
      enddo

    enddo
    enddo
    enddo
    enddo
  enddo

  call copy_boundary(y)

  return
end subroutine

subroutine conv_y_to_prop(y,this,ic0,is0)
!
! conversion from fermion field to quark propagator
!
!       y : fermion field
!    this : quark propagator
! ic0,is0 : source color-spinor of the quark propagator to be converted from fermion field.
!
  implicit none
  type(quark_prop),     intent(inout) :: this
  type(field_quark_wg), intent(in)    :: y
  integer, intent(in) :: ic0,is0
  integer :: ieoxyz,ith,ic,is,ieo,isp
  integer :: ix,iy,iz,it

  do ieo =0,1
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,ith,it,isp,is,ic)
    do ix=1,NX  ! site loop in 1 block.
    do iy=1,NY
    do iz=1,NZ
    ieoxyz=mod(ipeo+iz+iy+ix+ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      it=ith*2+ieoxyz
      isp=ispace(ix,iy,iz)

      do is=1,SPIN
      do ic=1,COL
        this%prop(ic,ic0,is,is0,isp,it)=y%eo(ieo)%s(ith,iz,iy,ix)%y(ic,is)
      enddo
      enddo

    enddo
    enddo
    enddo
    enddo
  enddo

  return
end subroutine

subroutine store_qprop_params(this,iout)
!
! store quark propagator on a file (parallel files)
!
  use error_class
  implicit none
  type(qprop_params), intent(inout) :: this
  integer,            intent(in) :: iout

  write(iout)this%quark_type

  select type(p => this%mass)
#ifdef _CLOVER
  type is (quark_clover)
#else
  type is (quark_wilson)
#endif

    write(iout)get_id(p)
    call save_config(p,iout)

  type is (quark_truncd_dw_overlap)

    write(iout)get_id(p)
    call save_config(p,iout)

  end select

  write(iout)this%fname
  write(iout)this%path
  write(iout)this%tol

  return
end subroutine

subroutine restore_qprop_params(this,iout)
!
! restore quark propagator on a file (parallel files)
!
  use error_class
  implicit none
  type(qprop_params), intent(inout) :: this
  integer,               intent(in) :: iout
  character(len=CHARLEN) :: str
  integer :: id,nflavor

  read(iout)this%quark_type

  select case(this%quark_type)
  case (0)
#ifdef _CLOVER
    allocate(quark_clover::this%mass)
#else
    allocate(quark_wilson::this%mass)
#endif

  case (1)

    allocate(quark_truncd_dw_overlap::this%mass)

  case default

    write(str,'("Quark type must be 0(Wilson/Clover) or 1(DWOVLP).")')
    call error_stop(str)

  end select

  read(iout)id

  select type(p => this%mass)
#ifdef _CLOVER
  type is (quark_clover)
#else
  type is (quark_wilson)
#endif

    call new(p,id)
    call read_config(p,iout)

  type is (quark_truncd_dw_overlap)

    call new(p,id)
    call read_config(p,iout)

  end select

  read(iout)this%fname
  read(iout)this%path
  read(iout)this%tol

  return
end subroutine

subroutine store_qprop(this,plq,path,head)
!
! store quark propagator on parallel files
!
  use file_tools_class
  use print_status_class
  use error_class
  implicit none
  type(quark_prop), intent(inout) :: this
  real(DP),         intent(in)    :: plq
  character(len=*), intent(in)    :: path,head
  integer :: iout,i
  character(len=CHARLEN) :: str,fname

  if ( 0 == nodeid ) then
    write(*,'(80("="))')
    write(*,'("    Store quark propagetor")')
  endif
  call print(this%parameter)

  write(str,'(".x",z1,"y",z1,"z",z1)')(ipsite(i),i=1,NDIM-1)
  fname = TRIM(TRIM(ADJUSTL(path))//'/'//TRIM(ADJUSTL(head)))//TRIM(ADJUSTL(str))

  write(str,'("    Store quark propagetor : ",A)')TRIM(fname)
  call print_status(str)

  iout = search_free_file_unit()
  open(iout,file=TRIM(ADJUSTL(fname)),status='new',form='unformatted')

  if (ALLOCATED(this%prop)) then
    write(iout)this%prop
  else
    write(str,'("quark propagator is not computed. ",A,I11)')__FILE__,__LINE__
    call error_stop(str)
  endif

  write(iout)NTX,NTY,NTZ,NTT
  write(iout)NDIMX,NDIMY,NDIMZ
  write(iout)nodeid
  write(iout)plq

  this%plq = plq

  write(iout)this%psum

  if (ASSOCIATED(this%parameter)) then
    call store(this%parameter,iout)
  else
    write(str,'("quark propagator physics parameter is not assigned. ",A,I11)')__FILE__,__LINE__
    call error_stop(str)
  endif

  if (ASSOCIATED(this%wavefunc_sous)) then
    call store(this%wavefunc_sous,iout)
  else
    write(str,'("quark propagator source wavefunction is not assigned. ",A,I11)')__FILE__,__LINE__
    call error_stop(str)
  endif

  if (ASSOCIATED(this%wavefunc_sink)) then
    call store(this%wavefunc_sink,iout)
  else
    write(str,'("quark propagator sink wavefunction is not assigned. ",A,I11)')__FILE__,__LINE__
    call error_stop(str)
  endif

  write(iout)this%smear_sous
  write(iout)this%smear_sink

  close(iout)

  if ( 0 == nodeid ) then
    write(*,'(80("="))')
  endif

  return
end subroutine

subroutine restore_qprop(this,plq,path,head)
!
! restore quark propagator from parallel files
!
  use file_tools_class
  use print_status_class
  use error_class
  implicit none
  type(quark_prop), intent(inout) :: this
  real(DP),         intent(out)   :: plq
  character(len=*), intent(in) :: path,head
  character(len=CHARLEN) :: str,fname
  real(DP), parameter :: tol = 1.0e-12_DP
  real(DP) :: psum0(NTT),diffp,dpsum
  integer :: iout,i,itt
  integer :: NTX0,NTY0,NTZ0,NTT0
  integer :: NDIMX0,NDIMY0,NDIMZ0
  integer :: nodeid0

  if ( 0 == nodeid ) then
    write(*,'(80("="))')
    write(*,'("  Restore quark propagetor")')
  endif

  call new(this%parameter%etime)
  call tic(this%parameter%etime)

  write(str,'(".x",z1,"y",z1,"z",z1)')(ipsite(i),i=1,NDIM-1)
  fname = TRIM(TRIM(ADJUSTL(path))//'/'//TRIM(ADJUSTL(head)))//TRIM(ADJUSTL(str))

  write(str,'("  Restore quark propagetor : ",A)')TRIM(fname)
  call print_status(str)

  iout = search_free_file_unit()
  open(iout,file=TRIM(ADJUSTL(fname)),status='old',form='unformatted')

  if (.not.ALLOCATED(this%prop)) then
    allocate(this%prop(COL,COL,SPIN,SPIN,NX*NY*NZ,NTT))
  endif
  read(iout)this%prop

  read(iout)NTX0,NTY0,NTZ0,NTT0
  read(iout)NDIMX0,NDIMY0,NDIMZ0
  read(iout)nodeid0
  read(iout)plq

  this%plq = plq

  if ( (NTX /= NTX0) .or. &
 &     (NTY /= NTY0) .or. &
 &     (NTZ /= NTZ0) .or. &
 &     (NTT /= NTT0) ) then
    write(str,'("Lattice size is not matched. (NTX,NTY,NTZ,NTT)=",4I3)')NTX0,NTY0,NTZ0,NTT0
    call error_stop(str)
  endif

  if ( (NDIMX /= NDIMX0) .or. &
 &     (NDIMY /= NDIMY0) .or. &
 &     (NDIMZ /= NDIMZ0) ) then
    write(str,'("Parallel size is not matched. (NDIMX,NDIMY,NDIMZ)=",3I3)')NDIMX0,NDIMY0,NDIMZ0
    call error_stop(str)
  endif

  if ( nodeid /= nodeid0 ) then
    write(str,'("nodeid is not matched. nodeid=",3I3)')nodeid
    call error_stop(str)
  endif

  read(iout)psum0


  if (ASSOCIATED(this%parameter)) then
    call restore(this%parameter,iout)
  else
    write(str,'("quark propagator physics parameter is not assigned. ",A,I11)')__FILE__,__LINE__
    call error_stop(str)
  endif

  if (ASSOCIATED(this%wavefunc_sous)) then
    call restore(this%wavefunc_sous,iout)
  else
    write(str,'("quark propagator source wavefunction is not assigned. ",A,I11)')__FILE__,__LINE__
    call error_stop(str)
  endif

  if (ASSOCIATED(this%wavefunc_sink)) then
    call restore(this%wavefunc_sink,iout)
  else
    write(str,'("quark propagator sink wavefunction is not assigned. ",A,I11)')__FILE__,__LINE__
    call error_stop(str)
  endif

  read(iout)this%smear_sous
  read(iout)this%smear_sink

  close(iout)

  call psum(this)

  if (0 == nodeid) then
    write(*,'(80("-"))')
    write(*,'("  Checksum PSUM")')
    write(*,'("              ORG                COMPUTED              DIFF")')
  endif

  dpsum = 0.0_DP

  do itt=1,NTT

    diffp = psum0(itt)-this%psum(itt)
    dpsum =dpsum + ABS(diffp)**2

    if (0 == nodeid) write(*,'(I4,3E24.15)')itt,psum0(itt),this%psum(itt),diffp
  enddo
  dpsum = SQRT(dpsum)/NTT

  if ( dpsum > tol ) then
    if (0 == nodeid) write(*,'(" DIFFSUM :",E24.15," >  TOL :",E24.15)')dpsum, tol
    write(str,'("quark propagator check sum error. ",A,I11)')__FILE__,__LINE__
    call error_stop(str)
  else
    if (0 == nodeid) write(*,'(" DIFFSUM :",E24.15," <= TOL :",E24.15)')dpsum, tol
  endif

  call print(this)

  if ( 0 == nodeid ) then
    write(*,'(80("="))')
  endif

  return
end subroutine


subroutine get_qprop(this,u)
!
! compute quark propagator
!
  use error_class
  implicit none
  type(quark_prop),      intent(inout) :: this
  type(vfield_gluon_wg), intent(in)    :: u
  type(timer) :: solver_time
#ifdef _USE_BLOCK_SOLVER
  type(field_quark_wg), allocatable :: y(:),Dy(:)
  integer :: ic,is,ics
  real(DP), allocatable :: rtmp0(:),rtmp1(:)
#else
  type(field_quark_wg), allocatable :: y,Dy
  integer :: ic,is
  real(DP) :: rtmp0,rtmp1
#endif
  character(len=CHARLEN) :: str
  integer :: maxiter,mode,iter
  integer :: iout
  real(DP) :: tol,etime
  real(DP) :: copy_y_time0,copy_y_time1,copy_y_etime
#ifdef _CLOVER
  type(tfield_gluon_wg), allocatable :: wl34,wl98
  real(DP) :: dummy
#endif
  type(dwf_low_modes),   allocatable :: LV
  integer :: itraj

  call new(this%parameter%etime)
  call tic(this%parameter%etime)

  itraj = this%itraj

  select type(p => this%parameter%mass)
  type is (quark_truncd_dw_overlap)
    if (p%is_lowmode_file) then
      allocate(LV)
      call restore_low_modes(LV,p,itraj,p%lowmode_path,p%lowmode_head)
    endif
  end select

#ifdef _CLOVER
!=======================
! compute clover term
!=======================
  select type(p => this%parameter%mass)
  type is (quark_clover)
    allocate(wl34,wl98)
    call make_two_links(u,wl34,wl98)
    call make_clover_leaf(u,wl34,wl98)
    deallocate(wl34,wl98)
    call make_inverse_clover_term(p,0,dummy)
    call delete_clover_leaf()
  end select
#endif

  if (.not.allocated(this%prop)) then
    call error_stop("get_qprop: Error, source is not setted/allocated.")
  endif

#ifdef _USE_BLOCK_SOLVER
!===============================
!  Blocked solver 
!===============================
  call new(solver_time)
  call tic(solver_time)

  copy_y_time0 = get_elapse(copy_fq_time)

  write(str,'("# BLOCK_SOLVER FOR COL=1,2,3 SPIN=1,2,3,4")')
  call print(this%parameter%solver_log,TRIM(str))
  iout = get_file_unit(this%parameter%solver_log)

  allocate(y(CLSP),Dy(CLSP))
  allocate(rtmp0(CLSP),rtmp1(CLSP))
  do ics=1,CLSP
    call new( y(ics))
    call new(dy(ics))
  enddo

  !------------------------
  ! Dy = D \ y
  !------------------------
  tol    = this%parameter%tol
  do is=1,SPIN
  do ic=1,COL
    ics = ic + (is-1)*COL
    call conv_prop_to_y(this, y(ics),ic,is)
    rtmp0(ics) = abs2(y(ics))
  enddo
  enddo

  select type(p => this%parameter%mass)
#ifdef _CLOVER
  type is (quark_clover)
#else
  type is (quark_wilson)
#endif
    call assign_inv_blk_mult_wd(p,iout,tol,iter,Dy,y,u)

  type is (quark_truncd_dw_overlap)

    write(str,'("Block solver for Truncated Overlap Fermions is not yet implemented.")')
    call error_stop(str)

  end select

  do is=1,SPIN
  do ic=1,COL
    ics = ic + (is-1)*COL
    rtmp1(ics) = abs2(Dy(ics))
    call conv_y_to_prop(Dy(ics),this,ic,is)
  enddo
  enddo

  copy_y_time1 = get_elapse(copy_fq_time)
  copy_y_etime = copy_y_time1-copy_y_time0

  call toc(solver_time)
  etime = get_elapse(solver_time)
  if (nodeid==0) then
    write(*,'(" block solver done  ETime:",E24.16," COPY_Y_ETime:",E24.16)')etime,copy_y_etime
    do is=1,SPIN
    do ic=1,COL
      ics = ic + (is-1)*COL
      write(*,'(" (col,spin)=",2I3," |y|^2:",E24.16," |Dy|^2:",E24.16)')ic,is,rtmp0(ics),rtmp1(ics)
    enddo
    enddo
  endif
  deallocate(rtmp0,rtmp1)

#else

!===============================
!  Non-Blocked solver 
!===============================
  allocate(y,Dy)
  do is=1,SPIN
  do ic=1,COL

    call new(solver_time)
    call tic(solver_time)

    copy_y_time0 = get_elapse(copy_fq_time)

    write(str,'("# SOLVER FOR COL=",I2," SPIN=",I2)')ic,is
    call print(this%parameter%solver_log,TRIM(str))
    iout = get_file_unit(this%parameter%solver_log)

    call new(y)
    call new(Dy)

    !------------------------
    ! Dy = D \ y
    !------------------------
    tol    = this%parameter%tol
    call conv_prop_to_y(this, y,ic,is)
    rtmp0 = abs2(y)
    select type(p => this%parameter%mass)
#ifdef _CLOVER
    type is (quark_clover)
#else
    type is (quark_wilson)
#endif

      call assign_inv_mult_wd(p,iout,tol,iter,Dy,y,u)

    type is (quark_truncd_dw_overlap)

      iter = 10000
      if (p%is_lowmode_file) then
        call assign_inv_mult_ov(p,LV,iout,tol,iter,Dy,y,u)
      else
        call assign_inv_mult_ov(p,   iout,tol,iter,Dy,y,u)
      endif

    end select
    rtmp1 = abs2(Dy)
    call conv_y_to_prop(Dy,this,ic,is)

    copy_y_time1 = get_elapse(copy_fq_time)
    copy_y_etime = copy_y_time1-copy_y_time0

    call toc(solver_time)
    etime = get_elapse(solver_time)
    if (nodeid==0) then
      write(*,'(" solver done ",2I3," ETime:",E24.16," COPY_Y_ETime:",E24.16," |y|^2:",E24.16," |Dy|^2:",E24.16)')&
&     ic,is,etime,copy_y_etime,rtmp0,rtmp1
    endif
  enddo
  enddo
#endif

  deallocate(y,Dy)
#ifdef _CLOVER
  select type(p => this%parameter%mass)
  type is (quark_clover)
    call delete_inverse_clover_term(p)
  end select
#endif

  select type(p => this%parameter%mass)
  type is (quark_truncd_dw_overlap)
    if (p%is_lowmode_file) then
      call delete(LV)
      deallocate(LV)
    endif
  end select

  call toc(this%parameter%etime)

  return
end subroutine

subroutine assign_inv_mult_wd(param,iout,tol,iiter,dy,y,u)
!
! assign a fermion fieldsmultiplied by inverse Wilson-Clover operator
!
!    dy := D \ y
!
!   param : quark parameters
!    iout : solver status/error io-unit
!     tol : solver tolerance
!   iiter : solver iteration count
!      dy : fermion field multiplied by a inverse Wilson-Clover operator
!       y : input fermion field
!       u : gauge link field
!
  use hmc_std_quark_wilson_class, only : assign_inv_mult_eoprec_wd
  implicit none
#ifdef _CLOVER
  type(quark_clover), intent(inout) :: param
#else
  type(quark_wilson), intent(inout) :: param
#endif
  integer,  intent(in)    :: iout
  real(DP), intent(in)    :: tol
  integer,  intent(inout) :: iiter
  type(field_quark_wg),  intent(in)    :: y
  type(field_quark_wg),  intent(inout) :: dy
  type(vfield_gluon_wg), intent(in)    :: u
  type(field_quark_wg),  allocatable :: w
#ifdef _DDD_
  type(field_quark_wg),  allocatable :: s
#endif
  real(DP) :: rtmp,rtmp2
  integer :: ieo,ioe
  ieo = 0
  ioe = 1-ieo

  allocate(w)
  call new(w)
#ifdef _CLOVER
  call assign_mult_clover_term(w%eo(ieo),param,y%eo(ieo))
  call assign_mult_clover_term(w%eo(ioe),param,y%eo(ioe))
#else
  call assign(w,y) ! w = y
#endif

#ifdef _DDD_
rtmp = abs2(w)
if(nodeid==0) write(*,'(" |F\y|^2:",E24.16)')rtmp
#endif

  !
  !  we = we + kappa Meo yo
  !
  call assign_mult_hop(param,dy%eo(ieo),w%eo(ioe),u)
  rtmp = get_kappa(param)
  call accum_add_mult(w%eo(ieo),dy%eo(ieo),rtmp)

#ifdef _DDD_
write(*,*)"kappa",rtmp
rtmp = abs2(w%eo(ieo))
if(nodeid==0) write(*,'(" |we|^2:",E24.16)')rtmp
#endif

  call assign_inv_mult_eoprec_wd(param,iout,tol,iiter,dy%eo(ieo),w%eo(ieo),u)

#ifdef _DDD_
rtmp = abs2(dy%eo(ieo))
if(nodeid==0) write(*,'(" |dye|^2:",E24.16)')rtmp
#endif

  !
  !  dyo = wo + kappa Moe dye
  !
  call assign(dy%eo(ioe),w%eo(ioe))
  call assign_mult_hop(param,w%eo(ioe),dy%eo(ieo),u)
  rtmp = get_kappa(param)
  call accum_add_mult(dy%eo(ioe),w%eo(ioe),rtmp)   ! dy = dy + w * rtmp


#ifdef _DDD_
rtmp = abs2(dy)
if(nodeid==0) write(*,'(" |dy|^2:",E24.16)')rtmp
!====================
! s = F\M dy
! w = dy
!====================
allocate(s)
call new(s)
call assign(w,dy)
call assign_mult_hop(param,s%eo(ieo),dy%eo(ioe),u)
call assign_mult_hop(param,s%eo(ioe),dy%eo(ieo),u)
rtmp = - get_kappa(param)
call accum_add_mult(w,s,rtmp)  ! w = w - s*kappa = F\D dy
!==============
! s = F\y - w
!==============
call assign_mult_clover_term(s%eo(ieo),param,y%eo(ieo))
call assign_mult_clover_term(s%eo(ioe),param,y%eo(ioe))
rtmp = abs2(s)
call accum_sub(s,w)
rtmp2 = abs2(s)
if(nodeid==0) write(*,'(" |F\(b-Dy)|/|F\b|:",E24.16)')sqrt(rtmp2/rtmp)
deallocate(s)
#endif

  deallocate(w)

!  call comlib_finalize()
!  stop
  return
end subroutine

subroutine assign_mult_u_q(uq,q,u,mu)
!  
! assign a quark propagtor shifted and multiplied by gauge link
!  
!  uq(n) := u_mu(n) q(n+mu) 
!  
!  uq : quark prop * link
!   q : quark prop
!   u : gauge link 
!  mu : direction
!  
  implicit none
  integer, intent(in) :: mu
  type(vfield_gluon_wg), intent(in)    :: u
  type(quark_prop),      intent(in)    :: q
  type(quark_prop),      intent(inout) :: uq
  type(field_quark_wg), allocatable :: y,uy
  integer :: ic,is

  call assign(uq,q)
  allocate(y,uy)
  call new(y)
  call new(uy)

  do is=1,SPIN
  do ic=1,COL
    call conv_prop_to_y(uq, y,ic,is)
    call assign_mult_u_y(uy,y,u,mu)
    call conv_y_to_prop(uy,uq,ic,is)
  enddo
  enddo

  deallocate(y,uy)

  return
end subroutine

subroutine assign_mult_u_y(uy,y,u,mu)
!  
! assign a fermion field shifted and multiplied by gauge link
!  
!  uy(n) := u_mu(n) y(n+mu) 
!  
!  uy : quark field * link
!   y : quark field
!   u : gauge link 
!  mu : direction
!  
  implicit none
  type(field_quark_wg),  intent(inout) :: uy, y
  type(vfield_gluon_wg), intent(in)    :: u
  integer, intent(in) :: mu

  call assign_mult_u_eo(uy%eo(0),y%eo(1),u%eo(0),mu)
  call assign_mult_u_eo(uy%eo(1),y%eo(0),u%eo(1),mu)

  return
end subroutine

subroutine assign_mult_u_eo(uye,yo,ue,mu)
!  
! assign a fermion field shifted and multiplied by gauge link (even/odd sites only)
!  
!  uye(n) := ue_mu(n) yo(n+mu) 
!  
!  uye : quark field * link (even/odd)
!   yo : quark field        (odd/even)
!   ue : gauge link         (even)
!   mu : direction
!  
!  
  use error_class
  implicit none
  type(field_quark_eo_wg),  intent(inout) :: uye,yo
  type(vfield_gluon_eo_wg), intent(in) :: ue
  integer, intent(in) :: mu

  integer :: ix,iy,iz,ith,ieoxyz,ic,is
  integer :: ix2,iy2,iz2,ith2,ieo,ioe

  ioe = yo%ieo
  ieo = ue%ieo
  if (ieo == ioe) then
    call error_stop("assign_mult_u_eo yo%ieo = ue%ieo ieo error.")
  endif
  uye%ieo = ue%ieo

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ix2,iy2,iz2,ieoxyz,ith,ith2,ic,is)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ

    ix2=ix+isx(mu)
    iy2=iy+isy(mu)
    iz2=iz+isz(mu)

    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      ith2=ith+ieoxyz*ist(mu)
      do is=1,SPIN
      do ic=1,COL
        uye%s(ith,iz,iy,ix)%y(ic,is) = ue%mu(mu)%s(ith ,iz ,iy ,ix )%u(ic,1)  &
 &                                           *yo%s(ith2,iz2,iy2,ix2)%y(1,is)  &
 &                                   + ue%mu(mu)%s(ith ,iz ,iy ,ix )%u(ic,2)  &
 &                                           *yo%s(ith2,iz2,iy2,ix2)%y(2,is)  &
 &                                   + ue%mu(mu)%s(ith ,iz ,iy ,ix )%u(ic,3)  &
 &                                           *yo%s(ith2,iz2,iy2,ix2)%y(3,is)
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine assign_inv_blk_mult_wd(param,iout,tol,iiter,dy,y,u)
!
! assign fermion fields multiplied by inverse Wilson-Clover operator (blocked version)
!
!    dy(:) := D \ y(:)
!
!   param : quark parameters
!    iout : solver status/error io-unit
!     tol : solver tolerance
!   iiter : solver iteration count
!      dy : fermion fields  multiplied by a inverse Wilson-Clover operator
!       y : input fermion fields
!       u : gauge link field
!
  use error_class
#ifdef _USE_MEASURE_CG_SOLVER
  use hmc_blk_quark_wilson_class, only : assign_inv_mult_Oee
#else
  use hmc_blk_quark_wilson_class, only : assign_inv_mult_eoprec_wd
#endif
  implicit none
#ifdef _CLOVER
  type(quark_clover), intent(inout) :: param
#else
  type(quark_wilson), intent(inout) :: param
#endif
  integer,  intent(in)    :: iout
  real(DP), intent(in)    :: tol
  integer,  intent(inout) :: iiter
  type(field_quark_wg),  intent(in)    ::  y(:)
  type(field_quark_wg),  intent(inout) :: dy(:)
  type(vfield_gluon_wg), intent(in)    :: u
  type(field_quark_wg),     allocatable :: w(:)
  type(field_quark_eo_wg),  allocatable :: we(:),dwe(:)
  real(DP) :: rtmp,rtmp2
  integer :: nblk,nblk_dy,iblk
  integer :: ieo,ioe
  ieo = 0
  ioe = 1-ieo

  nblk    = SIZE(y)
  nblk_dy = SIZE(dy)
  if ( nblk /= nblk_dy ) then
    call error_stop("assign_inv_blk_mult_wd : block size is inconsitent, SIZE(dy) /= SIZE(y).")
  endif

  allocate(w(nblk),we(nblk),dwe(nblk))
  do iblk=1,nblk
    call new(w(iblk))
#ifdef _CLOVER
    call assign_mult_clover_term(w(iblk)%eo(ieo),param,y(iblk)%eo(ieo))
    call assign_mult_clover_term(w(iblk)%eo(ioe),param,y(iblk)%eo(ioe))
#else
    call assign(w(iblk),y(iblk)) ! w = y
#endif

    !
    !  we = we + kappa Meo yo
    !
    call assign_mult_hop(param,dy(iblk)%eo(ieo),w(iblk)%eo(ioe),u)
    rtmp = get_kappa(param)
    call accum_add_mult(w(iblk)%eo(ieo),dy(iblk)%eo(ieo),rtmp)
    call assign(we(iblk),w(iblk)%eo(ieo))
  enddo

#ifdef _USE_MEASURE_CG_SOLVER
  call assign_inv_mult_Oee(param,iout,tol,iiter,dwe,we,u)
#else
  call assign_inv_mult_eoprec_wd(param,iout,tol,iiter,dwe,we,u)
#endif

  do iblk=1,nblk

#ifdef _USE_MEASURE_CG_SOLVER
    call assign_mult_eoprec_wd_dag(param,dy(iblk)%eo(ieo),dwe(iblk),u)
#else
    call assign(dy(iblk)%eo(ieo),dwe(iblk))
#endif

    !
    !  dyo = wo + kappa Moe dye
    !
    call assign(dy(iblk)%eo(ioe),w(iblk)%eo(ioe))
    call assign_mult_hop(param,w(iblk)%eo(ioe),dy(iblk)%eo(ieo),u)
    rtmp = get_kappa(param)
    call accum_add_mult(dy(iblk)%eo(ioe),w(iblk)%eo(ioe),rtmp)   ! dy = dy + w * rtmp
  enddo

  deallocate(w,we,dwe)

  return
end subroutine

end module
